# Example: Using routes to build a component
# A directional coupler is constructed by defining two routes
# with different waveguide definitions.

from ipkiss3 import all as i3

TECH = i3.TECH


class BentDirectionalCoupler(i3.PCell):

    """a directional coupler with bends on each side"""

    wg_template = i3.WaveguideTemplateProperty(
        default=i3.TECH.PCELLS.WG.DEFAULT, doc="waveguide template used by the directional coupler"
    )
    coupler_length = i3.PositiveNumberProperty(
        default=i3.TECH.WG.SHORT_STRAIGHT, doc="length of the directional coupler"
    )

    wg_north = i3.ChildCellProperty(doc="North arm cell")
    wg_south = i3.ChildCellProperty(doc="South arm cell")

    # define the wg pcells
    def _default_wg_north(self):
        wg_north = i3.RoundedWaveguide(name=self.name + "_wgnorth", trace_template=self.wg_template)
        return wg_north

    def _default_wg_south(self):
        wg_south = i3.RoundedWaveguide(name=self.name + "_wgsouth", trace_template=self.wg_template)
        return wg_south

    class Layout(i3.LayoutView):

        coupler_spacing = i3.PositiveNumberProperty(
            default=i3.TECH.WG.DC_SPACING, doc="spacing between the two waveguide centerlines"
        )
        bend_angle = i3.AngleProperty(default=45.0, doc="angle at which the directional coupler is bent")
        straight_after_bend = i3.PositiveNumberProperty(
            default=i3.TECH.WG.SHORT_STRAIGHT, doc="length of the straight waveguide after the bend"
        )
        bend_radius = i3.PositiveNumberProperty(
            default=i3.TECH.WG.BEND_RADIUS, doc="Bend_radius of the waveguides used"
        )

        def _default_wg_north(self):

            # 1. straight routes defined using points
            r1s = i3.RouteWithPorts(
                points=[(0, 0.5 * self.coupler_spacing), (self.coupler_length, 0.5 * self.coupler_spacing)],
                bend_radius=self.bend_radius,
            )

            # 2. define a section on each end to bend towards the correct angle
            r1_start = i3.RouteToAngle(
                start_port=r1s.ports[0],
                angle_out=180.0 - self.bend_angle,
                start_straight=0.0,
                end_straight=self.straight_after_bend,
                bend_radius=self.bend_radius,
            )

            r1_end = i3.RouteToAngle(
                start_port=r1s.ports[1],
                angle_out=self.bend_angle,
                start_straight=0.0,
                end_straight=self.straight_after_bend,
                bend_radius=self.bend_radius,
            )

            # 3. making the upper waveguide.
            r1 = r1_start.reversed() + r1s + r1_end

            wg_north_layout = self.cell.wg_north.get_default_view(i3.LayoutView)
            wg_north_layout.set(trace_template=self.wg_template, bend_radius=self.bend_radius, shape=r1)

            return wg_north_layout

        def _default_wg_south(self):
            r2s = i3.RouteWithPorts(
                points=[(0, -0.5 * self.coupler_spacing), (self.coupler_length, -0.5 * self.coupler_spacing)],
                bend_radius=self.bend_radius,
            )

            r2_start = i3.RouteToAngle(
                start_port=r2s.ports[0],
                angle_out=180.0 + self.bend_angle,
                start_straight=0.0,
                end_straight=self.straight_after_bend,
                bend_radius=self.bend_radius,
            )

            r2_end = i3.RouteToAngle(
                start_port=r2s.ports[1],
                angle_out=-self.bend_angle,
                start_straight=0.0,
                end_straight=self.straight_after_bend,
                bend_radius=self.bend_radius,
            )

            # add the pieces together
            r2 = r2_start.reversed() + r2s + r2_end

            wg_south_layout = self.cell.wg_south.get_default_view(i3.LayoutView)
            wg_south_layout.set(trace_template=self.wg_template, bend_radius=self.bend_radius, shape=r2)

            return wg_south_layout

        def _generate_instances(self, insts):
            insts += i3.SRef(reference=self.wg_north, name="w1")
            insts += i3.SRef(reference=self.wg_south, name="w2")
            return insts

        # Generating the ports of the directional coupler
        def _generate_ports(self, prts):
            # We reuse the  ports from the waveguides and change their names
            insts = self.instances
            prts += i3.expose_ports(
                insts,
                {
                    "w1:in": "in_1",
                    "w1:out": "out_1",
                    "w2:in": "in_2",
                    "w2:out": "out_2",
                },
            )

            return prts


if __name__ == "__main__":
    print("This is not the main file. Run 'execute.py' in the same folder")
